@@ -43,7 +43,7 @@
  * @author    Mike Pultz <mike@mikepultz.com>
  * @copyright 2010 Mike Pultz <mike@mikepultz.com>
  * @license   http://www.opensource.org/licenses/bsd-license.php  BSD License
- * @version   SVN: $Id: DNS2.php 216 2013-10-28 04:24:17Z mike.pultz $
+ * @version   SVN: $Id$
  * @link      http://pear.php.net/package/Net_DNS2
  * @since     File available since Release 0.6.0
  *
@@ -72,7 +72,7 @@ class Net_DNS2
     /*
      * the current version of this library
      */
-    const VERSION = '1.3.2';
+    const VERSION = '1.4.1';
 
     /*
      * the default path to a resolv.conf file
@@ -128,7 +128,7 @@ class Net_DNS2
     /*
      * the max size of the cache file (in bytes)
      */
-    public $cache_size = 10000;
+    public $cache_size = 50000;
 
     /*
      * the method to use for storing cache data; either "serialize" or "json"
@@ -203,23 +203,36 @@ class Net_DNS2
 
     /*
      * the EDNS(0) UDP payload size to use when making DNSSEC requests
-     * see RFC 2671 section 6.2.3 for more details
+     * see RFC 4035 section 4.1 - EDNS Support.
      *
-     * http://tools.ietf.org/html/draft-ietf-dnsext-rfc2671bis-edns0-10
+     * there is some different ideas on the suggest size to supprt; but it seems to
+     * be "at least 1220 bytes, but SHOULD support 4000 bytes.
+     *
+     * we'll just support 4000
      *
      */
-    public $dnssec_payload_size = 1280;
+    public $dnssec_payload_size = 4000;
+
+    /*
+     * the last exeception that was generated
+     */
+    public $last_exception = null;
+
+    /*
+     * the list of exceptions by name server
+     */
+    public $last_exception_list = array();
+
+    /*
+     * name server list
+     */
+    public $nameservers = array();
 
     /*
      * local sockets
      */
     protected $sock = array('udp' => array(), 'tcp' => array());
 
-    /*
-     * name server list
-     */
-    protected $nameservers = array();
-
     /*
      * if the socket extension is loaded
      */
@@ -240,11 +253,6 @@ class Net_DNS2
      */
     protected $use_cache = false;
 
-    /*
-     * the last erro message returned by the sockets class
-     */
-    private $_last_socket_error = '';
-
     /**
      * Constructor - base constructor for the Resolver and Updater
      *
@@ -257,9 +265,17 @@ class Net_DNS2
     public function __construct(array $options = null)
     {
         //
-        // check for the sockets extension
+        // check for the sockets extension; we no longer support the sockets library under 
+        // windows- there have been too many errors related to sockets under windows- 
+        // specifically inconsistent socket defines between versions of windows- 
         //
-        $this->sockets_enabled = extension_loaded('sockets');
+        // and since I can't seem to find a way to get the actual windows version, it 
+        // doesn't seem fixable in the code.
+        //
+        if ( (extension_loaded('sockets') == true) && (strtoupper(substr(PHP_OS, 0, 3)) !== 'WIN') ) {
+
+            $this->sockets_enabled = true;
+        }
 
         //
         // load any options that were provided
@@ -360,6 +376,14 @@ class Net_DNS2
 
         } else {
 
+            //
+            // temporary list of name servers; do it this way rather than just 
+            // resetting the local nameservers value, just incase an exception 
+            // is thrown here; this way we might avoid ending up with an empty 
+            // namservers list.
+            //
+            $ns = array();
+
             //
             // check to see if the file is readable
             //
@@ -404,7 +428,7 @@ class Net_DNS2
                             || (self::isIPv6($value) == true)
                         ) {
 
-                            $this->nameservers[] = $value;
+                            $ns[] = $value;
                         } else {
 
                             throw new Net_DNS2_Exception(
@@ -443,8 +467,21 @@ class Net_DNS2
                     Net_DNS2_Lookups::E_NS_INVALID_FILE
                 );
             }
+
+            //
+            // store the name servers locally
+            //
+            if (count($ns) > 0) {
+                $this->nameservers = $ns;
+            }
         }
 
+        //
+        // remove any duplicates; not sure if we should bother with this- if people
+        // put duplicate name servers, who I am to stop them?
+        //
+        $this->nameservers = array_unique($this->nameservers);
+
         //
         // check the name servers
         //
@@ -606,11 +643,14 @@ class Net_DNS2
         }
 
         //
-        // only RSAMD5 and RSASHA1 are supported for SIG(0)
+        // only RSA algorithms are supported for SIG(0)
         //
         switch($this->auth_signature->algorithm) {
         case Net_DNS2_Lookups::DNSSEC_ALGORITHM_RSAMD5:
         case Net_DNS2_Lookups::DNSSEC_ALGORITHM_RSASHA1:
+        case Net_DNS2_Lookups::DNSSEC_ALGORITHM_RSASHA256:
+        case Net_DNS2_Lookups::DNSSEC_ALGORITHM_RSASHA512:
+        case Net_DNS2_Lookups::DNSSEC_ALGORITHM_DSA:
             break;
         default:
             throw new Net_DNS2_Exception(
@@ -685,9 +725,7 @@ class Net_DNS2
         //
         if (extension_loaded('filter') == true) {
 
-            if (filter_var(
-                $_address, FILTER_VALIDATE_IP, FILTER_FLAG_IPV4
-            ) == false) {
+            if (filter_var($_address, FILTER_VALIDATE_IP, FILTER_FLAG_IPV4) == false) {
                 return false;
             }
         } else {
@@ -702,10 +740,7 @@ class Net_DNS2
             //
             // then make sure we're not a IPv6 address
             //
-            if (preg_match(
-                '/^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$/', 
-                $_address
-            ) == 0) {
+            if (preg_match('/^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$/', $_address) == 0) {
                 return false;
             }
         }
@@ -728,9 +763,7 @@ class Net_DNS2
         // use filter_var() if it's available; it's faster than preg
         //
         if (extension_loaded('filter') == true) {
-            if (filter_var(
-                $_address, FILTER_VALIDATE_IP, FILTER_FLAG_IPV6
-            ) == false) {
+            if (filter_var($_address, FILTER_VALIDATE_IP, FILTER_FLAG_IPV6) == false) {
                 return false;
             }
         } else {
@@ -745,9 +778,7 @@ class Net_DNS2
             //
             // then make sure it doesn't match a IPv4 address
             //
-            if (preg_match(
-                '/^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$/', $_address
-            ) == 1) {
+            if (preg_match('/^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$/', $_address) == 1) {
                 return false;
             }
         }
@@ -848,372 +879,471 @@ class Net_DNS2
         //
         $response = null;
         $ns = '';
-        $socket_type = null;
-        $tcp_fallback = false;
 
         while (1) {
 
             //
             // grab the next DNS server
             //
-            if ($tcp_fallback == false) {
+            $ns = each($this->nameservers);
+            if ($ns === false) {
 
-                $ns = each($this->nameservers);
-                if ($ns === false) {
+                if (is_null($this->last_exception) == false) {
+
+                    throw $this->last_exception;
+                } else {
 
                     throw new Net_DNS2_Exception(
-                        'every name server provided has failed: ' . 
-                        $this->_last_socket_error,
+                        'every name server provided has failed',
                         Net_DNS2_Lookups::E_NS_FAILED
                     );
                 }
+            }
 
-                $ns = $ns[1];
+            $ns = $ns[1];
+
+            //
+            // if the use TCP flag (force TCP) is set, or the packet is bigger than our 
+            // max allowed UDP size- which is either 512, or if this is DNSSEC request,
+            // then whatever the configured dnssec_payload_size is.
+            //
+            $max_udp_size = Net_DNS2_Lookups::DNS_MAX_UDP_SIZE;
+            if ($this->dnssec == true)
+            {
+                $max_udp_size = $this->dnssec_payload_size;
+            }
+
+            if ( ($use_tcp == true) || (strlen($data) > $max_udp_size) ) {
+
+                try
+                {
+                    $response = $this->sendTCPRequest($ns, $data, ($request->question[0]->qtype == 'AXFR') ? true : false);
+
+                } catch(Net_DNS2_Exception $e) {
+
+                    $this->last_exception = $e;
+                    $this->last_exception_list[$ns] = $e;
+
+                    continue;
+                }
+
+            //
+            // otherwise, send it using UDP
+            //
+            } else {
+
+                try
+                {
+                    $response = $this->sendUDPRequest($ns, $data);
+
+                    //
+                    // check the packet header for a trucated bit; if it was truncated,
+                    // then re-send the request as TCP.
+                    //
+                    if ($response->header->tc == 1) {
+
+                        $response = $this->sendTCPRequest($ns, $data);
+                    }
+
+                } catch(Net_DNS2_Exception $e) {
+
+                    $this->last_exception = $e;
+                    $this->last_exception_list[$ns] = $e;
+
+                    continue;
+                }
             }
 
             //
-            // if the use TCP flag (force TCP) is set, or the packet is bigger 
-            // than 512 bytes, use TCP for sending the packet
+            // make sure header id's match between the request and response
             //
-            if ( ($use_tcp == true) 
-                || (strlen($data) > Net_DNS2_Lookups::DNS_MAX_UDP_SIZE) 
-                || ($tcp_fallback == true)
-            ) {
-                $tcp_fallback = false;
-                $socket_type = Net_DNS2_Socket::SOCK_STREAM;
+            if ($request->header->id != $response->header->id) {
+
+                $this->last_exception = new Net_DNS2_Exception(
+
+                    'invalid header: the request and response id do not match.',
+                    Net_DNS2_Lookups::E_HEADER_INVALID,
+                    null,
+                    $request,
+                    $response
+                );
+
+                $this->last_exception_list[$ns] = $this->last_exception;
+                continue;
+            }
+
+            //
+            // make sure the response is actually a response
+            // 
+            // 0 = query, 1 = response
+            //
+            if ($response->header->qr != Net_DNS2_Lookups::QR_RESPONSE) {
+            
+                $this->last_exception = new Net_DNS2_Exception(
+
+                    'invalid header: the response provided is not a response packet.',
+                    Net_DNS2_Lookups::E_HEADER_INVALID,
+                    null,
+                    $request,
+                    $response
+                );
+
+                $this->last_exception_list[$ns] = $this->last_exception;
+                continue;
+            }
+
+            //
+            // make sure the response code in the header is ok
+            //
+            if ($response->header->rcode != Net_DNS2_Lookups::RCODE_NOERROR) {
+            
+                $this->last_exception = new Net_DNS2_Exception(
+                
+                    'DNS request failed: ' . 
+                    Net_DNS2_Lookups::$result_code_messages[$response->header->rcode],
+                    $response->header->rcode,
+                    null,
+                    $request,
+                    $response
+                );
+
+                $this->last_exception_list[$ns] = $this->last_exception;
+                continue;
+            }
+
+            break;
+        }
+
+        return $response;
+    }
+
+    /**
+     * sends a DNS request using TCP
+     *
+     * @param string  $_ns   the name server to use for the request
+     * @param string  $_data the raw DNS packet data
+     * @param boolean $_axfr if this is a zone transfer request
+     *
+     * @return Net_DNS2_Packet_Response the reponse object
+     * @throws Net_DNS2_Exception
+     * @access private
+     *
+     */
+    private function sendTCPRequest($_ns, $_data, $_axfr = false)
+    {
+        //
+        // grab the start time
+        //
+        $start_time = microtime(true);
+
+        //
+        // see if we already have an open socket from a previous request; if so, try to use
+        // that instead of opening a new one.
+        //
+        if ( (!isset($this->sock['tcp'][$_ns]))
+            || (!($this->sock['tcp'][$_ns] instanceof Net_DNS2_Socket))
+        ) {
+
+            //
+            // if the socket library is available, then use that
+            //
+            if ($this->sockets_enabled === true) {
+
+                $this->sock['tcp'][$_ns] = new Net_DNS2_Socket_Sockets(
+                    Net_DNS2_Socket::SOCK_STREAM, $_ns, $this->dns_port, $this->timeout
+                );
+
+            //
+            // otherwise the streams library
+            //
+            } else {
+
+                $this->sock['tcp'][$_ns] = new Net_DNS2_Socket_Streams(
+                    Net_DNS2_Socket::SOCK_STREAM, $_ns, $this->dns_port, $this->timeout
+                );
+            }
+
+            //
+            // if a local IP address / port is set, then add it
+            //
+            if (strlen($this->local_host) > 0) {
+
+                $this->sock['tcp'][$_ns]->bindAddress(
+                    $this->local_host, $this->local_port
+                );
+            }
+
+            //
+            // open the socket
+            //
+            if ($this->sock['tcp'][$_ns]->open() === false) {
+
+                throw new Net_DNS2_Exception(
+                    $this->sock['tcp'][$_ns]->last_error, Net_DNS2_Lookups::E_NS_SOCKET_FAILED
+                );
+            }
+        }
+
+        //
+        // write the data to the socket; if it fails, continue on
+        // the while loop
+        //
+        if ($this->sock['tcp'][$_ns]->write($_data) === false) {
+
+            throw new Net_DNS2_Exception(
+                $this->sock['tcp'][$_ns]->last_error, Net_DNS2_Lookups::E_NS_SOCKET_FAILED
+            );
+        }
+
+        //
+        // read the content, using select to wait for a response
+        //
+        $size = 0;
+        $result = null;
+        $response = null;
+
+        //
+        // handle zone transfer requests differently than other requests.
+        //
+        if ($_axfr == true) {
+
+            $soa_count = 0;
+
+            while (1) {
 
                 //
-                // create the socket object
+                // read the data off the socket
                 //
-                if ( (!isset($this->sock['tcp'][$ns])) 
-                    || (!($this->sock['tcp'][$ns] instanceof Net_DNS2_Socket)) 
-                ) {
-                    if ($this->sockets_enabled === true) {
+                $result = $this->sock['tcp'][$_ns]->read($size, ($this->dnssec == true) ? $this->dnssec_payload_size : Net_DNS2_Lookups::DNS_MAX_UDP_SIZE);
+                if ( ($result === false) || ($size < Net_DNS2_Lookups::DNS_HEADER_SIZE) ) {
 
-                        $this->sock['tcp'][$ns] = new Net_DNS2_Socket_Sockets(
-                            Net_DNS2_Socket::SOCK_STREAM, 
-                            $ns, 
-                            $this->dns_port, 
-                            $this->timeout
-                        );
-                    } else {
-
-                        $this->sock['tcp'][$ns] = new Net_DNS2_Socket_Streams(
-                            Net_DNS2_Socket::SOCK_STREAM, 
-                            $ns, 
-                            $this->dns_port, 
-                            $this->timeout
-                        );
-                    }
-                }
-
-                //
-                // if a local IP address / port is set, then add it
-                //
-                if (strlen($this->local_host) > 0) {
-
-                    $this->sock['tcp'][$ns]->bindAddress(
-                        $this->local_host, $this->local_port
+                    throw new Net_DNS2_Exception(
+                        $this->sock['tcp'][$_ns]->last_error, Net_DNS2_Lookups::E_NS_SOCKET_FAILED
                     );
                 }
 
                 //
-                // open it; if it fails, continue in the while loop
+                // parse the first chunk as a packet
                 //
-                if ($this->sock['tcp'][$ns]->open() === false) {
-
-                    $this->_last_socket_error = $this->sock['tcp'][$ns]->last_error;
-                    continue;
-                }
+                $chunk = new Net_DNS2_Packet_Response($result, $size);
 
                 //
-                // write the data to the socket; if it fails, continue on 
-                // the while loop
+                // if this is the first packet, then clone it directly, then
+                // go through it to see if there are two SOA records
+                // (indicating that it's the only packet)
                 //
-                if ($this->sock['tcp'][$ns]->write($data) === false) {
+                if (is_null($response) == true) {
 
-                    $this->_last_socket_error = $this->sock['tcp'][$ns]->last_error;
-                    continue;
-                }
+                    $response = clone $chunk;
 
-                //
-                // read the content, using select to wait for a response
-                //
-                $size = 0;
-                $result = null;
+                    //
+                    // look for a failed response; if the zone transfer
+                    // failed, then we don't need to do anything else at this
+                    // point, and we should just break out.                 
+                    //
+                    if ($response->header->rcode != Net_DNS2_Lookups::RCODE_NOERROR) {
+                        break;
+                    }
 
-                //
-                // handle zone transfer requests differently than other requests.
-                //
-                if ($request->question[0]->qtype == 'AXFR') {
-
-                    $soa_count = 0;
-
-                    while (1) {
+                    //   
+                    // go through each answer
+                    //
+                    foreach ($response->answer as $index => $rr) {
 
                         //
-                        // read the data off the socket
+                        // count the SOA records
                         //
-                        $result = $this->sock['tcp'][$ns]->read($size);
-                        if ( ($result === false) 
-                            ||  ($size < Net_DNS2_Lookups::DNS_HEADER_SIZE) 
-                        ) {
-                            $this->_last_socket_error = $this->sock['tcp'][$ns]->last_error;
-                            break;
+                        if ($rr->type == 'SOA') {
+                            $soa_count++;
                         }
+                    }
 
-                        //
-                        // parse the first chunk as a packet
-                        //
-                        $chunk = new Net_DNS2_Packet_Response($result, $size);
-
-                        //
-                        // if this is the first packet, then clone it directly, then
-                        // go through it to see if there are two SOA records 
-                        // (indicating that it's the only packet)
-                        //
-                        if (is_null($response) == true) {
-
-                            $response = clone $chunk;
-
-                            //
-                            // look for a failed response; if the zone transfer
-                            // failed, then we don't need to do anything else at this
-                            // point, and we should just break out.
-                            //
-                            if ($response->header->rcode != Net_DNS2_Lookups::RCODE_NOERROR) {
-                                break;
-                            }
-
-                            //
-                            // go through each answer
-                            //
-                            foreach ($response->answer as $index => $rr) {
-
-                                //
-                                // count the SOA records
-                                //
-                                if ($rr->type == 'SOA') {
-                                    $soa_count++;
-                                }
-                            }
-
-                            //
-                            // if we have 2 or more SOA records, then we're done; 
-                            // otherwise continue out so we read the rest of the 
-                            // packets off the socket
-                            //
-                            if ($soa_count >= 2) {
-                                break;
-                            } else {
-                                continue;
-                            }
-
-                        } else {
-
-                            //
-                            // go through all these answers, and look for SOA records
-                            //
-                            foreach ($chunk->answer as $index => $rr) {
-
-                                //
-                                // count the number of SOA records we find
-                                //
-                                if ($rr->type == 'SOA') {
-                                    $soa_count++;
-                                }
-
-                                //
-                                // add the records to a single response object
-                                //
-                                $response->answer[] = $rr;
-                            }
-
-                            //
-                            // if we've found the second SOA record, we're done
-                            //
-                            if ($soa_count >= 2) {
-                                break;
-                            }
-                        }
+                    //
+                    // if we have 2 or more SOA records, then we're done;
+                    // otherwise continue out so we read the rest of the 
+                    // packets off the socket
+                    //
+                    if ($soa_count >= 2) {
+                        break;
+                    } else {
+                        continue;
                     }
 
                 } else {
 
-                    $result = $this->sock['tcp'][$ns]->read($size);
-                    if ( ($result === false) 
-                        ||  ($size < Net_DNS2_Lookups::DNS_HEADER_SIZE) 
-                    ) {
-                        $this->_last_socket_error = $this->sock['tcp'][$ns]->last_error;
-                        continue;
+                    //
+                    // go through all these answers, and look for SOA records
+                    //
+                    foreach ($chunk->answer as $index => $rr) {
+
+                        //
+                        // count the number of SOA records we find
+                        //
+                        if ($rr->type == 'SOA') {
+                            $soa_count++;           
+                        }
+
+                        //
+                        // add the records to a single response object
+                        //
+                        $response->answer[] = $rr;                  
                     }
 
                     //
-                    // create the packet object
+                    // if we've found the second SOA record, we're done
                     //
-                    $response = new Net_DNS2_Packet_Response($result, $size);
-                }
-
-                break;
-
-            } else {
-
-                $socket_type = Net_DNS2_Socket::SOCK_DGRAM;
-
-                //
-                // create the socket object
-                //
-                if ( (!isset($this->sock['udp'][$ns])) 
-                    || (!($this->sock['udp'][$ns] instanceof Net_DNS2_Socket))
-                ) {
-                    if ($this->sockets_enabled === true) {
-
-                        $this->sock['udp'][$ns] = new Net_DNS2_Socket_Sockets(
-                            Net_DNS2_Socket::SOCK_DGRAM, $ns, $this->dns_port, $this->timeout
-                        );
-                    } else {
-
-                        $this->sock['udp'][$ns] = new Net_DNS2_Socket_Streams(
-                            Net_DNS2_Socket::SOCK_DGRAM, $ns, $this->dns_port, $this->timeout
-                        );
+                    if ($soa_count >= 2) {
+                        break;
                     }
-                }            
-
-                //
-                // if a local IP address / port is set, then add it
-                //
-                if (strlen($this->local_host) > 0) {
-
-                    $this->sock['udp'][$ns]->bindAddress(
-                        $this->local_host, $this->local_port
-                    );
                 }
-
-                //
-                // open it
-                //
-                if ($this->sock['udp'][$ns]->open() === false) {
-
-                    $this->_last_socket_error = $this->sock['udp'][$ns]->last_error;
-                    continue;
-                }
-
-                //
-                // write the data to the socket
-                //
-                if ($this->sock['udp'][$ns]->write($data) === false) {
-
-                    $this->_last_socket_error = $this->sock['udp'][$ns]->last_error;
-                    continue;
-                }
-
-                //
-                // read the content, using select to wait for a response
-                //
-                $size = 0;
-
-                $result = $this->sock['udp'][$ns]->read($size);
-                if (( $result === false) 
-                    || ($size < Net_DNS2_Lookups::DNS_HEADER_SIZE)
-                ) {
-                    $this->_last_socket_error = $this->sock['udp'][$ns]->last_error;
-                    continue;
-                }
-
-                //
-                // create the packet object
-                //
-                $response = new Net_DNS2_Packet_Response($result, $size);
-                if (is_null($response)) {
-
-                    throw new Net_DNS2_Exception(
-                        'empty response object', 
-                        Net_DNS2_Lookups::E_NS_FAILED,
-                        null,
-                        $request
-                    );
-                }
-
-                //
-                // check the packet header for a trucated bit; if it was truncated,
-                // then re-send the request as TCP.
-                //
-                if ($response->header->tc == 1) {
-
-                    $tcp_fallback = true;
-                    continue;
-                }
-
-                break;
             }
+
+        //
+        // everything other than a AXFR
+        //
+        } else {
+
+            $result = $this->sock['tcp'][$_ns]->read($size, ($this->dnssec == true) ? $this->dnssec_payload_size : Net_DNS2_Lookups::DNS_MAX_UDP_SIZE);
+            if ( ($result === false) || ($size < Net_DNS2_Lookups::DNS_HEADER_SIZE) ) {
+
+                throw new Net_DNS2_Exception(
+                    $this->sock['tcp'][$_ns]->last_error, Net_DNS2_Lookups::E_NS_SOCKET_FAILED
+                );
+            }
+
+            //
+            // create the packet object
+            //
+            $response = new Net_DNS2_Packet_Response($result, $size);
         }
 
         //
-        // if $response is null, then we didn't even try once; which shouldn't
-        // actually ever happen
+        // store the query time
         //
-        if (is_null($response)) {
-
-            throw new Net_DNS2_Exception(
-                'empty response object', 
-                Net_DNS2_Lookups::E_NS_FAILED,
-                null,
-                $request
-            );
-        }
+        $response->response_time = microtime(true) - $start_time;
 
         //
         // add the name server that the response came from to the response object,
         // and the socket type that was used.
         //
-        $response->answer_from = $ns;
-        $response->answer_socket_type = $socket_type;
+        $response->answer_from = $_ns;
+        $response->answer_socket_type = Net_DNS2_Socket::SOCK_STREAM;
 
         //
-        // make sure header id's match between the request and response
+        // return the Net_DNS2_Packet_Response object
         //
-        if ($request->header->id != $response->header->id) {
+        return $response;
+    }
+
+    /**
+     * sends a DNS request using UDP
+     *
+     * @param string  $_ns   the name server to use for the request
+     * @param string  $_data the raw DNS packet data
+     *
+     * @return Net_DNS2_Packet_Response the reponse object
+     * @throws Net_DNS2_Exception
+     * @access private
+     *
+     */
+    private function sendUDPRequest($_ns, $_data)
+    {
+        //
+        // grab the start time
+        //
+        $start_time = microtime(true);
+
+        //
+        // see if we already have an open socket from a previous request; if so, try to use
+        // that instead of opening a new one.
+        //
+        if ( (!isset($this->sock['udp'][$_ns]))
+            || (!($this->sock['udp'][$_ns] instanceof Net_DNS2_Socket))
+        ) {
+
+            //
+            // if the socket library is available, then use that
+            //
+            if ($this->sockets_enabled === true) {
+
+                $this->sock['udp'][$_ns] = new Net_DNS2_Socket_Sockets(
+                    Net_DNS2_Socket::SOCK_DGRAM, $_ns, $this->dns_port, $this->timeout
+                );
+
+            //
+            // otherwise the streams library
+            //
+            } else {
+
+                $this->sock['udp'][$_ns] = new Net_DNS2_Socket_Streams(
+                    Net_DNS2_Socket::SOCK_DGRAM, $_ns, $this->dns_port, $this->timeout
+                );
+            }
+
+            //
+            // if a local IP address / port is set, then add it
+            //
+            if (strlen($this->local_host) > 0) {
+
+                $this->sock['udp'][$_ns]->bindAddress(
+                    $this->local_host, $this->local_port
+                );
+            }
+
+            //
+            // open the socket
+            //
+            if ($this->sock['udp'][$_ns]->open() === false) {
+
+                throw new Net_DNS2_Exception(
+                    $this->sock['udp'][$_ns]->last_error, Net_DNS2_Lookups::E_NS_SOCKET_FAILED
+                );
+            }
+        }
+
+        //
+        // write the data to the socket
+        //
+        if ($this->sock['udp'][$_ns]->write($_data) === false) {
 
             throw new Net_DNS2_Exception(
-                'invalid header: the request and response id do not match.',
-                Net_DNS2_Lookups::E_HEADER_INVALID,
-                null,
-                $request,
-                $response
+                $this->sock['udp'][$_ns]->last_error, Net_DNS2_Lookups::E_NS_SOCKET_FAILED
             );
         }
 
         //
-        // make sure the response is actually a response
-        // 
-        // 0 = query, 1 = response
+        // read the content, using select to wait for a response
         //
-        if ($response->header->qr != Net_DNS2_Lookups::QR_RESPONSE) {
+        $size = 0;
+
+        $result = $this->sock['udp'][$_ns]->read($size, ($this->dnssec == true) ? $this->dnssec_payload_size : Net_DNS2_Lookups::DNS_MAX_UDP_SIZE);
+        if (( $result === false) || ($size < Net_DNS2_Lookups::DNS_HEADER_SIZE)) {
 
             throw new Net_DNS2_Exception(
-                'invalid header: the response provided is not a response packet.',
-                Net_DNS2_Lookups::E_HEADER_INVALID,
-                null,
-                $request,
-                $response
+                $this->sock['udp'][$_ns]->last_error, Net_DNS2_Lookups::E_NS_SOCKET_FAILED
             );
         }
 
         //
-        // make sure the response code in the header is ok
+        // create the packet object
         //
-        if ($response->header->rcode != Net_DNS2_Lookups::RCODE_NOERROR) {
+        $response = new Net_DNS2_Packet_Response($result, $size);
 
-            throw new Net_DNS2_Exception(
-                'DNS request failed: ' . 
-                Net_DNS2_Lookups::$result_code_messages[$response->header->rcode],
-                $response->header->rcode,
-                null,
-                $request,
-                $response
-            );
-        }
+        //
+        // store the query time
+        //
+        $response->response_time = microtime(true) - $start_time;
 
+        //
+        // add the name server that the response came from to the response object,
+        // and the socket type that was used.
+        //
+        $response->answer_from = $_ns;
+        $response->answer_socket_type = Net_DNS2_Socket::SOCK_DGRAM;
+
+        //
+        // return the Net_DNS2_Packet_Response object
+        //
         return $response;
     }
 }
